//P.java - P(u,v) permutation block
//Copyright (c) 2007 Thomas Schneider
//See full Copyright license terms in file ../GPL.txt

package SFE.PF;

import SFE.PF.UC.X;

/**
 * P(u,v) permutation block.
 * <p>
 * Permute the u inputs to the v outputs.
 * <p>
 * Description in <a href="http://portal.acm.org/citation.cfm?id=321449">Abraham Waksman, "A Permutation Network"</a>.
 *
 * @author Thomas Schneider
 */
abstract public class P extends Block {
	X s1[], s2[];
	Block b1, b2;
	
	/**
	 * Constructor
	 * @param u #inputs
	 * @param v #outputs
	 * @param c UC containing this selection block
	 */
	public P(int u, int v, UC c) {
		super(u,v,c);
	}

	// ---------------------------------------------------
	
	void program_X(X[] s, int i, boolean v) {
		if(!debug) s[i].program(v);
		else System.out.println(((s==s1)?"S1":"S2")+"["+i+"]->"+(v?"X":"H")); // debug
	}	

	//double linked list; node n is head
	class TodoList {
		int nextu[];
		int prevu[];
		int n;
		
		void remove(int x) {
			nextu[prevu[x]]=nextu[x];
			prevu[nextu[x]]=prevu[x];
		}

		int next() {
			int ret=nextu[n];
			if(ret==n) return -1;
			else return ret;
		}
		
		TodoList(int n) {
			this.n=n;
			nextu=new int[n+1];
			prevu=new int[n+1];
			for(int i=0; i<n+1; i++) {
				nextu[i]=(i+1) % (n+1);
				prevu[i]=(i+n) % (n+1);
			}		
		}
	}
}
