//P_EP.java - P(u,v>=u) expanded permutation block
//Copyright (c) 2007 Thomas Schneider
//See full Copyright license terms in file ../GPL.txt

package SFE.PF;

import SFE.BOAL.Gate;
import SFE.PF.UC.X;

/**
 * P(u,v>=u) expanded permutation block.
 *
 * @author Thomas Schneider
 */
public class P_EP extends P {
	TodoList Todo;
	
	public Gate[] extractCircuit(Gate[] inputs, Gate[] outputs) {
		if(outputs==null) outputs=new Gate[v];
		assert(inputs.length==u && outputs.length==v);
		if (u==1) { // wire input to all outputs
			for(int i=0; i<v; i++) outputs[i]=inputs[0];
			return outputs;
		}
		else {
			Gate[] in_p1=new Gate[u/2];
			Gate[] in_p2=new Gate[u-(u/2)];

			// first row X
			Gate[] outtmp=new Gate[2];
			for(int i=0; i<s1.length; i++) {
				Gate[] out_s1=s1[i].extractCircuit(inputs[2*i],inputs[2*i+1],outtmp);
				in_p1[i]=out_s1[0];
				in_p2[i]=out_s1[1];
			}
			if(u%2==1) {
				in_p2[in_p2.length-1]=inputs[inputs.length-1];
			}
			
			Gate[] out_p1=b1.extractCircuit(in_p1,null);
			Gate[] out_p2=b2.extractCircuit(in_p2,null);

			// last row X
			for(int i=0; i<s2.length; i++) {
				Gate[] out_s2=s2[i].extractCircuit(out_p1[i],out_p2[i],outtmp);
				outputs[2*i]=out_s2[0];
				outputs[2*i+1]=out_s2[1];
			}
			if(v%2==1) {
				outputs[v-1]=out_p2[out_p2.length-1];
			}
			else if(u%2==0) {
				outputs[v-2]=out_p1[out_p1.length-1];				
				outputs[v-1]=out_p2[out_p2.length-1];
			}
			
			return outputs;
		}
	}
	
	private void program_rec(int in, int block, int p1[], int p2[], int rows[], int cols[]) {
		int out=rows[in];
		
		if((in^1) < u && rows[in^1] != -1) {
			program_X(s1,in/2,(block==0)!=(in%2==0));
			Todo.remove(in/2);
		}
		
		if(block==1) {
			p2[in/2]=out/2;
			if(out/2<s2.length) {
				program_X(s2,out/2,out%2==0);
			}
		}
		else { // block==0
			p1[in/2]=out/2;
			if(out/2<s2.length) {
				program_X(s2,out/2,out%2==1);
			}
		}
		rows[in]=-1; cols[out]=-1;
		
		int newout=out^1;
		if(newout<v && cols[newout]!=-1) {
			int newin=cols[newout];
			cols[newout]=-1;
			program_rec(newin,block^1,p1,p2,rows,cols);
		}
		
		if((in^1) < u && rows[in^1] != -1) {
			program_rec(in^1,block^1,p1,p2,rows,cols);
		}
	}
	
	/**
 	 * Program expanded permutation block
 	 * 
	 * @param p input mapping: output[p[i]] <- input[i], i=0..(u-1)
	 */
	public void program(int p[]) {
		assert(p.length==u);

		if(u==1) return;

		if(debug) {
			System.out.print("p=[");
			for(int i=0; i<u; i++) System.out.print((i!=0?", ":"")+p[i]);
			System.out.println("]");
		}
		
		int rows[]=p;
		
		int cols[]=new int[v];
		for(int i=0; i<v; i++) cols[i]=-1;
		for(int i=0; i<u; i++) {
			int x=p[i];
			assert(x>=0 && x<v && cols[x]==-1); // each input is mapped to a distinct output
			cols[x]=i;
		}

		// programs for sub-blocks
		int p1[]=new int[u/2];
		int p2[]=new int[u-(u/2)];
		
		Todo=new TodoList(u/2);
		if (u%2==1) { // case c+d
			program_rec(u-1, 1, p1, p2, rows, cols);
		}
		if (v%2==1) { // case b+d
			if(cols[v-1]!=-1) program_rec(cols[v-1], 1, p1, p2, rows, cols);
		}
		if (u%2==0 && v%2==0) { // case a
			if(cols[v-1]!=-1) program_rec(cols[v-1], 1, p1, p2, rows, cols);
			if(cols[v-2]!=-1) program_rec(cols[v-2], 0, p1, p2, rows, cols);
		}
		
		for(int n=Todo.next(); n!=-1; n=Todo.next()) {
			program_rec(2*n,0,p1,p2,rows,cols);
		}
		
		// program sub-blocks
		b1.program(p1);
		b2.program(p2);
	}
	
	public int countGates() {
		if(u==1) return 0;
		else return (s1.length+s2.length)*UC.Xgates+b1.countGates()+b2.countGates();
	}
	
	public static int estimateGates(int u, int v) {
		if(v<u) return Integer.MAX_VALUE;
		else if(u==1) return 0;
		else {
			int s1count=u/2;
			int s2count=(u%2==0 && v%2==0)?v/2-1:v/2;
			return (s1count+s2count)*UC.Xgates + P_EP.estimateGates(u/2,v/2) + P_EP.estimateGates(u-(u/2),v-(v/2));
		}
	}
	
	public P_EP(int u, int v, UC c) {
		super(u,v,c);
		assert(v>=u);
		if (u==1); // wires only
		else {
			// first row X
			s1=new X[u/2];
			if(c!=null) for(int i=0; i<u/2; i++) s1[i]=c.new X();

			// B1
			b1=new P_EP(u/2, v/2, c);

			// B2
			b2=new P_EP(u-(u/2), v-(v/2), c);

			// last row X
			{
				int s2count=(u%2==0 && v%2==0)?v/2-1:v/2;
				s2=new X[s2count];
				if(c!=null) for (int i=0; i<s2count; i++) s2[i]=c.new X();
			}
		}
	}
	
	// debug
	public static void main(String[] args) {
		P_EP t;
		
		for(int i=1; i<40; i++) {
			t=new P_EP(i,10000,null);
			System.out.println(i+": "+t.countGates()/2);
		}
		//t.program(new int[]{49,0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19});
	}
}
