//P_EP.java - P(u>=v,v) truncated permutation block
//Copyright (c) 2007 Thomas Schneider
//See full Copyright license terms in file ../GPL.txt

package SFE.PF;

import SFE.BOAL.Gate;
import SFE.PF.UC.X;

/**
 * P(u>=v,v) truncated permutation block.
 *
 * @author Thomas Schneider
 */
public class P_TP extends P {
	TodoList Todo;

	public Gate[] extractCircuit(Gate[] inputs, Gate[] outputs) {
		if(outputs==null) outputs=new Gate[v];
		assert(inputs.length==u && outputs.length==v);
		if (v==1) { // wire selection block to in- and outputs
			outputs=b1.extractCircuit(inputs, outputs);
			return outputs;
		}
		else {
			Gate[] in_p1=new Gate[u/2];
			Gate[] in_p2=new Gate[u-(u/2)];

			// first row X
			Gate[] outtmp=new Gate[2];
			for(int i=0; i<s1.length; i++) {
				Gate[] out_s1=s1[i].extractCircuit(inputs[2*i],inputs[2*i+1],outtmp);
				in_p1[i]=out_s1[0];
				in_p2[i]=out_s1[1];
			}
			if(u%2==1) {
				in_p2[in_p2.length-1]=inputs[inputs.length-1];
			}
			
			Gate[] out_p1=b1.extractCircuit(in_p1,null);
			Gate[] out_p2=b2.extractCircuit(in_p2,null);

			// last row X
			for(int i=0; i<s2.length; i++) {
				Gate[] out_s2=s2[i].extractCircuit(out_p1[i],out_p2[i],outtmp);
				outputs[2*i]=out_s2[0];
				outputs[2*i+1]=out_s2[1];
			}
			if(v%2==1) {
				outputs[v-1]=out_p2[out_p2.length-1];
			}
			else if(u%2==0) {
				outputs[v-2]=out_p1[out_p1.length-1];				
				outputs[v-1]=out_p2[out_p2.length-1];
			}
			
			return outputs;
		}
	}
	
	private void program_rec(int out, int block, int p1[], int p2[], int rows[], int cols[]) {
		int in=cols[out];
		
		if((out^1) < v && cols[out^1] != -1 && out/2 < s2.length) {
			program_X(s2,out/2,(block==0)!=(out%2==0));
			Todo.remove(out/2);
		}
		
		if(block==1) {
			p2[out/2]=in/2;
			if(in/2<s1.length) {
				program_X(s1,in/2,in%2==0);
			}
		}
		else { // block==0
			p1[out/2]=in/2;
			if(in/2<s1.length) {
				program_X(s1,in/2,in%2==1);
			}
		}
		cols[out]=-1; rows[in]=-1;
		
		int newin=in^1;
		if(newin<u && rows[newin]!=-1) {
			int newout=rows[newin];
			rows[newin]=-1;
			program_rec(newout,block^1,p1,p2,rows,cols);
		}
		
		if((out^1) < v && cols[out^1] != -1 && out/2 < s2.length) {
			program_rec(out^1,block^1,p1,p2,rows,cols);
		}
	}
	
	/**
	 * Program truncated permutation block
	 * 
	 * @param p output mapping: output[i] <- input[p[i]], i=0..(v-1)
	 */
	public void program(int p[]) {
		assert(p.length==v);

		if(v==1) {
			b1.program(p);
			return;
		}
		
		if(debug) {
			System.out.print("p=[");
			for(int i=0; i<v; i++) System.out.print((i!=0?", ":"")+p[i]);
			System.out.println("]");
		}
		
		int cols[]=p;
		
		int rows[]=new int[u];
		for(int i=0; i<u; i++) rows[i]=-1;
		for(int i=0; i<v; i++) {
			int x=p[i];
			assert(x>=0 && x<u && rows[x]==-1); // each output is mapped to a distinct input
			rows[x]=i;
		}

		// programs for sub-blocks
		int p1[]=new int[v/2];
		int p2[]=new int[v-(v/2)];
		
		Todo=new TodoList(s2.length);
		if (u%2==1) { // case c+d
			if(rows[u-1]!=-1) program_rec(rows[u-1], 1, p1, p2, rows, cols);
		}
		if (v%2==1) { // case b+d
			if(cols[v-1]!=-1) program_rec(v-1, 1, p1, p2, rows, cols);
		}
		if (u%2==0 && v%2==0) { // case a
			program_rec(v-1, 1, p1, p2, rows, cols);
			if(cols[v-2]!=-1) program_rec(v-2, 0, p1, p2, rows, cols);
		}
		
		for(int n=Todo.next(); n!=-1; n=Todo.next()) {
			program_rec(2*n,0,p1,p2,rows,cols);
		}
		
		// program sub-blocks
		b1.program(p1);
		b2.program(p2);
	}
	
	public int countGates() {
		if(v==1) return b1.countGates();
		else return (s1.length+s2.length)*UC.Xgates+b1.countGates()+b2.countGates();
	}

	public static int estimateGates(int u, int v) {
		if(u<v) return Integer.MAX_VALUE;
		else if(v==1) return (u-1)*UC.Ygates;
		else {
			int s1count=u/2;
			int s2count=(u%2==0 && v%2==0)?v/2-1:v/2;
			return (s1count+s2count)*UC.Xgates + P_TP.estimateGates(u/2,v/2) + P_TP.estimateGates(u-(u/2),v-(v/2));
		}
	}
	
	public P_TP(int u, int v, UC c) {
		super(u,v,c);
		assert(u>=v);
		if (v==1) { // selection block
			b1=new S_u_1(u,c);
		}
		else {
			// first row X
			s1=new X[u/2];
			if(c!=null) for(int i=0; i<u/2; i++) s1[i]=c.new X();

			// B1
			b1=new P_TP(u/2, v/2, c);

			// B2
			b2=new P_TP(u-(u/2), v-(v/2), c);

			// last row X
			{
				int s2count=(u%2==0 && v%2==0)?v/2-1:v/2;
				s2=new X[s2count];
				if(c!=null) for (int i=0; i<s2count; i++) s2[i]=c.new X();
			}
		}
	}
	
	// debug
	public static void main(String[] args) {
		P_TP t;
		
		t=new P_TP(4,3,null);
		t.program(new int[]{3,1,2});
		
		for(int i=32; i<150; i++) {
			t=new P_TP(i,32,null);
			System.out.println(i+": "+t.countGates());
		}
	}
}
