//S_U.java - S(u>=v,v) selection block
//Copyright (c) 2007 Thomas Schneider
//See full Copyright license terms in file ../GPL.txt

package SFE.PF;

import SFE.BOAL.Gate;
import SFE.PF.UC.Y;

/**
 * S(u>=v,v) selection block.
 *
 * @author Thomas Schneider
 */
public class S_U extends S {
	Block b1;
	Y[] g;
	P_EP b2;

	public void program(int[] s) {
		assert(s.length==v);
		
		// count number of occurrences
		int c[]=new int[u];
		for(int i=0; i<v; i++) {
			assert(s[i]>=0 && s[i]<u);
			c[s[i]]++;
		}
		
		int sum=0;
		int prog1[]=new int[v];
		int upos[]=new int[u];
		int dummy[]=new int[u]; int ndummys=0; // list of dummy inputs
		for(int i=0; i<u; i++) {
			if(c[i]>0) {
				prog1[sum]=i;
				upos[i]=sum;
				if(sum>0) {
					if(g[sum-1]!=null) g[sum-1].program(true);
					else System.out.println("Y["+(sum-1)+"]=R");
				}
				for(int j=sum+1; j<sum+c[i]; j++) {
					if(g[j-1]!=null) g[j-1].program(false);
					else System.out.println("Y["+(j-1)+"]=L");
					prog1[j]=-1;
				}
				sum+=c[i];
			}
			else dummy[ndummys++]=i;
		}
		
		// program dummys
		int nextdummy=0;
		for(int i=0; i<v; i++) {
			if(prog1[i]==-1) prog1[i]=dummy[nextdummy++];
		}
		b1.program(prog1);

		// program b2
		int prog2[]=new int[v];
		for(int i=0; i<v; i++) {
			prog2[i]=upos[s[i]]++;
		}
		b2.program(prog2);
	}
	
	public int countGates() {
		if(v==1) return b1.countGates();
		else return b1.countGates()+g.length*UC.Ygates+b2.countGates();
	}
	
	public int estimateGates(int u, int v) {
		if(u<v) return Integer.MAX_VALUE;
		else if(v==1) return (u-1)*UC.Ygates;
		else return P_TP.estimateGates(u,v)+(v-1)*UC.Ygates+P_EP.estimateGates(v,v);
	}

	public Gate[] extractCircuit(Gate[] inputs, Gate[] outputs) {
		if(outputs==null) outputs=new Gate[v];
		if(v==1) {
			return b1.extractCircuit(inputs, outputs);
		}
		else {
			Gate[] out_p1=b1.extractCircuit(inputs,null);

			Gate[] out_g=new Gate[v];
			out_g[0]=out_p1[0];
			for(int i=1; i<v; i++) {
				out_g[i]=g[i-1].extractCircuit(out_g[i-1],out_p1[i]);
			}			
			return b2.extractCircuit(out_g,outputs);
		}
	}

	public S_U(int u, int v, UC c) {
		super(u,v,c);
	}
	public S_U() {
		super();
	}
	
	public void generate(int u, int v, UC c) {
		assert(u>=v);
		this.u=u; this.v=v;
		if(v==1) { // S_u_1 selection block
			b1=new S_u_1(u,c);
		}
		else {
			b1=new P_TP(u,v,c);
			g=new Y[v-1]; // first is wire
			if(c!=null) {
				for(int i=0; i<v-1; i++) {
					g[i]=c.new Y();
				}
			}
			b2=new P_EP(v,v,c);
		}
	}
	
	// debug
	public static void main(String[] args) {
		S_U s;

		s=new S_U(128,4,null);
		System.out.println(s.countGates());
		s.program(new int[]{0,0,4,3});
	}
}
