//S_U.java - S(u,v>=u) selection block
//Copyright (c) 2007 Thomas Schneider
//See full Copyright license terms in file ../GPL.txt

package SFE.PF;

import SFE.BOAL.Gate;
import SFE.PF.UC.Y;

/**
 * S(u,v>=u) selection block.
 *
 * @author Thomas Schneider
 */
public class S_V extends S {
	P_EP b1;
	Y[] g;
	P_TP b2;
	
	public void program(int[] s) {
		int prog1[];
		int prog2[];
		assert(s.length==v);
		
		if(u==1) return;
		
		// count number of occurrences
		int c[]=new int[u];
		for(int i=0; i<v; i++) {
			assert(s[i]>=0 && s[i]<u);
			c[s[i]]++;
		}
		
		int sum=0;
		prog1=new int[u];
		int upos[]=new int[u];
		int dummy[]=new int[v]; int ndummys=0; // list of dummy outputs
		for(int i=0; i<u; i++) {
			if(c[i]>0) {
				prog1[i]=sum;
				upos[i]=sum;
				if(sum>0) {
					if(g[sum-1]!=null) g[sum-1].program(true);
					else System.out.println("Y["+(sum-1)+"]=R");
				}
				for(int j=sum+1; j<sum+c[i]; j++) {
					if(g[j-1]!=null) g[j-1].program(false);
					else System.out.println("Y["+(j-1)+"]=L");
					dummy[ndummys++]=j;
				}
				sum+=c[i];
			}
		}
		
		// program dummys
		int nextdummy=0;
		for(int i=0; i<u; i++) {
			if(c[i]==0) {
				prog1[i]=dummy[nextdummy++];
			}
		}
		b1.program(prog1);

		// program b2
		prog2=new int[v];
		for(int i=0; i<v; i++) {
			prog2[i]=upos[s[i]]++;
		}
		b2.program(prog2);
	}
	
	public int countGates() {
		if(u==1) return 0;
		else return b1.countGates()+g.length*UC.Ygates+b2.countGates();
	}

	public int estimateGates(int u, int v) {
		if(v<u) return Integer.MAX_VALUE;
		else if(u==1) return 0;
		else return P_EP.estimateGates(u,v)+(v-1)*UC.Ygates+P_TP.estimateGates(v,v);
	}
	
	public Gate[] extractCircuit(Gate[] inputs, Gate[] outputs) {
		Gate[] out_p1;
		Gate[] out_g;
		if(outputs==null) outputs=new Gate[v];
		if(u==1) {
			for(int i=0; i<v; i++) outputs[i]=inputs[0];
			return outputs;
		}
		else {
			out_p1=b1.extractCircuit(inputs,null);

			out_g=new Gate[v];
			out_g[0]=out_p1[0];
			for(int i=1; i<v; i++) {
				out_g[i]=g[i-1].extractCircuit(out_g[i-1],out_p1[i]);
			}			
			return b2.extractCircuit(out_g,outputs);
		}
	}

	public S_V(int u, int v, UC c) {
		super(u,v,c);
	}
	public S_V() {
		super();
	}
	
	public void generate(int u, int v, UC c) {
		assert(v>=u);
		this.u=u; this.v=v;
		if(u==1) ; // wires only
		else {
			b1=new P_EP(u,v,c);
			g=new Y[v-1]; // first is wire
			if(c!=null) {
				for(int i=1; i<v; i++) {
					g[i-1]=c.new Y();
				}
			}
			b2=new P_TP(v,v,c);
		}
	}
	
	// debug
	public static void main(String[] args) {
		S_V s;

		s=new S_V(1,3,null);
		System.out.println(s.countGates());
		s.program(new int[]{0,0,0});

	}
}
